/*ms----------------------------------------------------------------------------


   Module:        m_data2raster.c

   Purpose: To extract raster data from the velocity metadata

   Procedures:
      Depend        - To check program dependencies and verify needed program data.

      usage         - To print a usage message on stdout.

      getData       - Generate the data 

      ParseArgs     - parse the command line.

      swab_struct   - converts big endian to little endian 


   Author:      James Miller
 
----------------------------------------------------------------------------me*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <stdlib.h>

#if defined __linux__
#define FSEEK fseeko64
#elif defined __sgi
#define FSEEK fseek64
#elif defined __sun
#define FSEEK fseeko64
#else
#define FSEEK fseek
#endif


/* ---- Local data types ---- */

typedef struct {
  int   n_columns;
  int   n_rows;
  float ul_east;
  float ul_north;
  float sp_east;
  float sp_north;
  int   confl_size;
  char  conflation_filename[256];
  char  metadata_filename[256];
  char  index_filename[256];
  } MetaDataHeader;

typedef struct {                     /* command line options...       */
   int           debug;                     /* Was "-db" specified?          */
   int           depend;                    /* Was "-depend" specified?      */
   int           help;                      /* Was "-h" specified?           */
   int           ul_x;
   int           ul_y;
   int           bounds_given;
   int           size_x;
   int           size_y;
   int           field;
   int           swab_in;
   int           swab_out;
   char          in_file[256];
   char          out_file[256];
   MetaDataHeader *head;
} Options_t;


typedef struct {
  float elevation;
  float slope_x;
  float slope_y;
  float Vx;
  float Vy;
  float Vz;
  float speed_error;
  float dir_error;
  long int   offset;
}VelocityMetaDataPoint;

 enum {
  ELEVATION,
  SLOPE_X,
  SLOPE_Y,
  MAGNITUDE,
  VX,
  VY,
  VZ,
  SPEED_ERROR,
  DIR_ERROR,
  N_FIELDS };   


char *field_names[] = {
  "elevation",
  "slope_x",
  "slope_y",
  "magnitude",
  "Vx",
  "Vy",
  "Vz",
  "speed_error",
  "dir_error",
   };


int Depend( Options_t  *options);
void usage(char *cmd);
int getData( Options_t  *options);

int ParseArgs(
   int        argc,
   char       *argv[],
   Options_t  *options);

/*fs----------------------------------------------------------------------------

    Procedure:  main

    Purpose:    Get coherence value and associated data at an x, y coordinate

    Exits:      Exit status is 0 on success, 1 on failure

----------------------------------------------------------------------------fe*/

int main(int argc, char *argv[])
{
   Options_t *options;

   options = (Options_t *)calloc(1, sizeof(Options_t));

   if (ParseArgs(argc, argv, options)) {
      usage(argv[0]);
      exit(1);
      }

   if(options->help) {
      usage(argv[0]);
      exit(0);
      }

   if (Depend(options)) {
      exit(1);
      }

   if (getData(options)) {
      exit(1);
      }

   exit(0);
}

/*----------------------------------------------------------------------------

  Procedure:  getHead

  Purpose:    Allocate header structure and populate from parameter file.

---------------------------------------------------------------------------- */

MetaDataHeader *getHead(Options_t *options)
{
   MetaDataHeader *header;
   FILE           *fp;
   char            line[256], *strptr;

   fp = fopen(options->in_file, "r");
   if(fp == NULL) {
       printf("Failed to open %s to read\n", options->in_file);
       return NULL;
       }

   header = (MetaDataHeader *)calloc(1, sizeof(MetaDataHeader));
   if(header == NULL) {
      printf("Failed to allocate memory for header\n");
      return NULL;
      }

   while(fgets(line, 255, fp)) {
       if(strstr(line, "number_of_columns")) {
          strptr = strstr(line, ":");
          strptr++;
          sscanf(strptr, "%d", &header->n_columns);
          }
       if(strstr(line, "number_of_rows")) {
          strptr = strstr(line, ":");
          strptr++;
          sscanf(strptr, "%d", &header->n_rows);
          }
       if(strstr(line, "easting_spacing")) {
          strptr = strstr(line, ":");
          strptr++;
          sscanf(strptr, "%f", &header->sp_east);
          }
       if(strstr(line, "northing_spacing")) {
          strptr = strstr(line, ":");
          strptr++;
          sscanf(strptr, "%f", &header->sp_north);
          }
       if(strstr(line, "bytes_per_conflation_point")) {
          strptr = strstr(line, ":");
          strptr++;
          sscanf(strptr, "%d", &header->confl_size);
          }
       if(strstr(line, "metadata_filename")) {
          strptr = strstr(line, ":");
          strptr++;
          while(isspace(*strptr)) strptr++;
          sprintf(header->metadata_filename, "%s", strptr);
          strptr = header->metadata_filename + strlen(header->metadata_filename)-1;
          while(!isalpha(*(strptr-1))) strptr--;
          *strptr = '\0';
          }
       }
   fclose(fp);

   return header;
}

/*fs----------------------------------------------------------------------------

    Procedure:  int Depend

    Purpose:    To check program dependencies

                Ensures that all data needed is present and reports any
                missing necessary files.

    Arguments:  Options_t *options - Pointer to command line options.
                Input_t   *input   - Pointer to input structure.

    Returns:    Returns 0 on success or 1 on failure.

----------------------------------------------------------------------------fe*/


int Depend(Options_t *options)
{
  int missing_file = 0;
  MetaDataHeader *header;

  header = getHead(options);
  if(header == NULL) {
      printf("Missing header file: %s\n\n", options->in_file);
      return 1;
      }

  if(access(header->metadata_filename, 04) ) {
      printf("Missing File: %s\n\n", header->metadata_filename);
      missing_file = 1;
      }
   options->head = header;
   return missing_file;
}

/* swab 4 byte float buffer */

int swab_4bytes(char *ptr) 
{
    unsigned char *cp, tmp;

   /*
    * Swap two pairs of bytes:
    */

    cp = (unsigned char *)ptr;

    tmp   = cp[3];
    cp[3] = cp[0];
    cp[0] = tmp;

    tmp   = cp[2];
    cp[2] = cp[1];
    cp[1] = tmp;

return 0;
}

/* swab each element of the given structure */

void swab_struct(VelocityMetaDataPoint *point)
{
  swab_4bytes((char *)&point->elevation);
  swab_4bytes((char *)&point->slope_x);
  swab_4bytes((char *)&point->slope_y);
  swab_4bytes((char *)&point->Vx);
  swab_4bytes((char *)&point->Vy);
  swab_4bytes((char *)&point->Vz);
  swab_4bytes((char *)&point->speed_error);
  swab_4bytes((char *)&point->dir_error);
  swab_4bytes((char *)&point->offset);
}

void swab_structold(VelocityMetaDataPoint *point)
{
  float f_val;
  long int i_val;

  swab((char *)&point->elevation, (char *)&f_val, sizeof(float));
  point->elevation = f_val;

  swab((char *)&point->slope_x, (char *)&f_val, sizeof(float));
  point->slope_x = f_val;

  swab((char *)&point->slope_y, (char *)&f_val, sizeof(float));
  point->slope_y = f_val;

  swab((char *)&point->Vx, (char *)&f_val, sizeof(float));
  point->Vx = f_val;

  swab((char *)&point->Vy, (char *)&f_val, sizeof(float));
  point->Vy = f_val;

  swab((char *)&point->Vz, (char *)&f_val, sizeof(float));
  point->Vz = f_val;

  swab((char *)&point->speed_error, (char *)&f_val, sizeof(float));
  point->speed_error = f_val;

  swab((char *)&point->dir_error, (char *)&f_val, sizeof(float));
  point->dir_error = f_val;

  swab((char *)&point->offset, (char *)&i_val, sizeof(long int));
  point->offset = i_val;
}

/*fs----------------------------------------------------------------------------

    Procedure:  getData

    Purpose:    Extract data and write to raster file

    Arguments:  Options_t *options - Pointer to command line options.
                Input_t   *input   - Pointer to input structure.

    Returns:    Returns 0 on success or 1 on failure.

----------------------------------------------------------------------------fe*/

int getData(Options_t  *options)
{
  FILE *in_fp, *out_fp;
  long long int   offset;
  int            start_i, start_j, end_i, end_j;
  int             ii, jj, buf_size, count;
  VelocityMetaDataPoint *v_buf;
  float                 *f_buf;
  MetaDataHeader *header = options->head;

 /* 
  * The set bounds option hasn't been implemented in the 
  * parse arguments function.
*/

  if(options->bounds_given ) {
      start_j = (options->ul_x - header->ul_east) / header->sp_east;
      if(start_j < 0) start_j = 0;
      end_j   =  start_j + options->size_x / header->sp_east;
      if(end_j > header->n_rows - 1) end_j = header->n_rows - 1;
      start_i = (header->ul_north - options->ul_y) / header->sp_north;
      if(start_i < 0) start_i = 0;
      end_i   = start_i + options->size_y / header->sp_north;
      if(end_i > header->n_columns - 1) end_i = header->n_columns - 1;
      }
  else {
      start_i = start_j = 0;
      end_i = header->n_rows;
      end_j = header->n_columns;
      }

  buf_size = end_j - start_j;
  v_buf = (VelocityMetaDataPoint *)calloc(buf_size, sizeof(VelocityMetaDataPoint));

  in_fp = fopen(header->metadata_filename, "rb");

  out_fp = fopen(options->out_file, "wb");

  f_buf = (float *)calloc(buf_size, sizeof(float));

/*
 * For each row, simply read in the array of velocity points
 * and add the appropriate field element to the output buffer
 * the write that buffer to the output file.
 */

  for(ii = start_i; ii < end_i; ii++) {
     offset = ii * header->n_columns + start_j;
     offset *= sizeof(VelocityMetaDataPoint);
     if(FSEEK(in_fp, offset, SEEK_SET)) {
         printf("fseek  at line %d\n", ii);
         exit(1);
         }
     count = fread((void *)v_buf, sizeof(VelocityMetaDataPoint), buf_size, in_fp);
     if(count != buf_size) {
         printf("read %d but expected %d\n", count, buf_size);
         exit(1);
         }
     for(jj = start_j; jj < end_j; jj++) {
         if(options->swab_in) {
              swab_struct(&v_buf[jj]);
              }
         switch (options->field) {
            case ELEVATION:
               f_buf[jj] = v_buf[jj].elevation;
               break;
            case  SLOPE_X:
               f_buf[jj] = v_buf[jj].slope_x;
               break;
            case  SLOPE_Y:
               f_buf[jj] = v_buf[jj].slope_y;
               break;
            case  VX:
               f_buf[jj] = v_buf[jj].Vx;
               break;
            case  VY:
               f_buf[jj] = v_buf[jj].Vy;
               break;
            case  VZ:
               f_buf[jj] = v_buf[jj].Vz;
               break;
            case  MAGNITUDE:
               f_buf[jj] = (float)sqrt((double)(v_buf[jj].Vx * v_buf[jj].Vx) +
                                       (double)(v_buf[jj].Vy * v_buf[jj].Vy) +
                                       (double)(v_buf[jj].Vz * v_buf[jj].Vz) );
               break;
            case  SPEED_ERROR:
               f_buf[jj] = v_buf[jj].speed_error;
               break;
            case  DIR_ERROR:
               f_buf[jj] = v_buf[jj].dir_error;
               break;
            default:
               f_buf[jj] = 0.0;
            }
         if(options->swab_out) {
              swab_4bytes((char *)&f_buf[jj]);
              }
        }
    fwrite((void *)f_buf, sizeof(float), buf_size, out_fp);
    }

  fclose(in_fp);
  fclose(out_fp);

  return 0;
}

/*fs----------------------------------------------------------------------------

    Procedure:  ParseArgs

    Purpose:    Parses the command line arguments


    Arguments :
              int argc   number of arguments expected
              char argv  array of arguments
              Options_t *options - Pointer to command line options.

    Returns:    Returns 0 on success or 1 on failure.

----------------------------------------------------------------------------fe*/

int 
ParseArgs(int argc, char *argv[], Options_t  *options)
{
   int    ii, jj;

   options->help = options->debug = options->depend = 0;
   options->field = MAGNITUDE;

   for(ii = 1; ii < argc; ii++) {
      if(!strcmp(argv[ii], "-h")) {
         options->help = 1;
      } else if(!strcmp(argv[ii], "-db")) {
         options->debug = 1;
      } else if(!strcmp(argv[ii], "-swab_in")) {
         options->swab_in = 1;
      } else if(!strcmp(argv[ii], "-swab_out")) {
         options->swab_out = 1;
      } else if(!strcmp(argv[ii], "-depend")) {
         options->depend = 1;
      } else if(!strcmp(argv[ii], "-in")) {
         ii++; 
         strcpy(options->in_file, argv[ii]);
      } else if(!strcmp(argv[ii], "-out")) {
         ii++; 
         strcpy(options->out_file, argv[ii]);
      } else {
        for(jj = 0; jj < N_FIELDS; jj++) {
            if(strstr(argv[ii], field_names[jj])) break;
            }
        if( jj == N_FIELDS) {
           printf("%s not valid flag\n", argv[ii]);
           return 1;
           }
        options->field = jj;
        }
     }

  if(options->in_file[0] == '\0') {
      printf("missing input parameter file\n");
      return 1;
      }

  if(options->out_file[0] == '\0') {
      printf("missing output data file\n");
      return 1;
      }

   return 0;
}

/*fs----------------------------------------------------------------------------

    Procedure:  usage

    Purpose:    Print program usage to stdout


    Arguments:  char *cmd  - program name

    Returns:    nothing

----------------------------------------------------------------------------fe*/

void usage(char *cmd)
{
   int jj;
   printf("\n%s: extracts a user specified field of the \n", cmd);
   printf("metadata velocities into a floating point raster file\n");
   printf("specified by user.  The input parameter file contains\n");
   printf("the names of the data files, the upper left map coordinates,\n");
   printf("pixel spacing and the raster dimensions.  No output header\n");
   printf("file is created.  That will have to be created in whatever\n");
   printf("format that the user needs.  All needed information is contained\n");
   printf("in the input parameter file\n\n");
   printf("Usage: %s -in <input.vel_meta.par> -out <output raster> -field_name\n", cmd);
   printf("Field names are as  follows:\n");
   for(jj = 0; jj< N_FIELDS;jj++) 
      printf("                             -%s\n", field_names[jj]);
   printf("\n\n           [-swab_in]       -Swap bytes of input if necessary.\n");
   printf("           [-swab_out]      -Swap bytes of output if necessary.\n\n");
   printf("  The velocity data provided is in big endian order so if you are\n");
   printf("  extracting raster data on a little endian machine, use the -swab_in flag\n");
   printf("  If you are running this on a big endian machine for a little endian\n");
   printf("  target machine, use -swab_out\n\n");
   printf("           [-h]              - print usage\n");
}
